# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains the metric_tensor batch_transform which wraps multiple
methods of computing the metric tensor.
"""
import functools
import warnings

import pennylane as qml
from pennylane import numpy as np
from pennylane.circuit_graph import LayerData

from .batch_transform import batch_transform


def expand_fn(tape, approx=None, allow_nonunitary=True, aux_wire=None, device_wires=None):
    """Set the metric tensor based on whether non-unitary gates are allowed."""
    # pylint: disable=unused-argument,too-many-arguments
    if not allow_nonunitary and approx is None:  # pragma: no cover
        return qml.transforms.expand_nonunitary_gen(tape)
    return qml.transforms.expand_multipar(tape)


@functools.partial(batch_transform, expand_fn=expand_fn)
def metric_tensor(tape, approx=None, allow_nonunitary=True, aux_wire=None, device_wires=None):
    r"""Returns a function that computes the metric tensor of a given QNode or quantum tape.

    The metric tensor convention we employ here has the following form:

    .. math::

        \text{metric_tensor}_{i, j} = \text{Re}\left[ \langle \partial_i \psi(\bm{\theta}) | \partial_j \psi(\bm{\theta}) \rangle
        - \langle \partial_i \psi(\bm{\theta}) | \psi(\bm{\theta}) \rangle \langle \psi(\bm{\theta}) | \partial_j \psi(\bm{\theta}) \rangle \right]

    with short notation :math:`| \partial_j \psi(\bm{\theta}) \rangle := \frac{\partial}{\partial \theta_j}| \psi(\bm{\theta}) \rangle`.
    It is closely related to the quantum fisher information matrix, see :func:`~.pennylane.qinfo.transforms.quantum_fisher` and eq. (27) in `arxiv:2103.15191 <https://arxiv.org/abs/2103.15191>`_.

    .. note::

        Only gates that have a single parameter and define a ``generator`` are supported.
        All other parametrized gates will be decomposed if possible.

        The ``generator`` of all parametrized operations, with respect to which the
        tensor is computed, are assumed to be Hermitian.
        This is the case for unitary single-parameter operations.

    Args:
        tape (pennylane.QNode or .QuantumTape): quantum tape or QNode to find the metric tensor of
        approx (str): Which approximation of the metric tensor to compute.

            - If ``None``, the full metric tensor is computed

            - If ``"block-diag"``, the block-diagonal approximation is computed, reducing
              the number of evaluated circuits significantly.

            - If ``"diag"``, only the diagonal approximation is computed, slightly
              reducing the classical overhead but not the quantum resources
              (compared to ``"block-diag"``).

        allow_nonunitary (bool): Whether non-unitary operations are allowed in circuits
            created by the transform. Only relevant if ``approx`` is ``None``.
            Should be set to ``True`` if possible to reduce cost.
        aux_wire (int or str or pennylane.wires.Wires): Auxiliary wire to be used for
            Hadamard tests. If ``None`` (the default), a suitable wire is inferred
            from the (number of) used wires in the original circuit and ``device_wires``.
        device_wires (.wires.Wires): Wires of the device that is going to be used for the
            metric tensor. Facilitates finding a default for ``aux_wire`` if ``aux_wire``
            is ``None``.
        hybrid (bool): Specifies whether classical processing inside a QNode
            should be taken into account when transforming a QNode.

            - If ``True``, and classical processing is detected, the Jacobian of the classical
              processing will be computed and included. When evaluated, the
              returned metric tensor will be with respect to the QNode arguments.
              The output shape can vary widely.

            - If ``False``, any internal QNode classical processing will be
              **ignored**. When evaluated, the returned metric tensor will be with
              respect to the **gate** arguments, and not the QNode arguments.
              The output shape is a single two-dimensional tensor.

    Returns:
        function or tuple[list[QuantumTape], function]:

        - If the input is a QNode, an object representing the metric tensor (function) of the
          QNode that takes the same arguments as the QNode and can be executed to obtain the
          metric tensor (matrix).

        - If the input is a tape, a tuple containing a
          list of generated tapes, together with a post-processing
          function to be applied to the results of the evaluated tapes
          in order to obtain the metric tensor.

    The block-diagonal part of the metric tensor always is computed using the
    covariance-based approach. If no approximation is selected,
    the off block-diagonal is computed using Hadamard tests.

    .. warning::

        Performing the Hadamard tests requires a device
        that has an additional wire as compared to the wires on which the
        original circuit was defined. This wire may be specified via ``aux_wire``.
        The available wires on the device may be specified via ``device_wires``.

        By default (that is, if ``device_wires=None`` ), contiguous wire
        numbering and usage is assumed and the additional
        wire is set to the next wire of the device after the circuit wires.

        If the given or inferred ``aux_wire`` does not exist on the device,
        a warning is raised and the block-diagonal approximation is computed instead.
        It is significantly cheaper in this case to explicitly set ``approx="block-diag"`` .

    The flag ``allow_nonunitary`` should be set to ``True`` whenever the device with
    which the metric tensor is computed supports non-unitary operations.
    This will avoid additional decompositions of gates, in turn avoiding a potentially
    large number of additional Hadamard test circuits to be run.
    State vector simulators, for example, often allow applying operations that are
    not unitary.
    On a real QPU, setting this flag to ``True`` may cause exceptions because the
    computation of the metric tensor will request invalid operations on a quantum
    device.

    **Example**

    Consider the following QNode:

    .. code-block:: python

        dev = qml.device("default.qubit", wires=3)

        @qml.qnode(dev, interface="autograd")
        def circuit(weights):
            qml.RX(weights[0], wires=0)
            qml.RY(weights[1], wires=0)
            qml.CNOT(wires=[0, 1])
            qml.RZ(weights[2], wires=1)
            qml.RZ(weights[3], wires=0)
            return qml.expval(qml.PauliZ(0) @ qml.PauliZ(1)), qml.expval(qml.PauliY(1))

    We can use the ``metric_tensor`` transform to generate a new function that returns the
    metric tensor of this QNode:

    >>> mt_fn = qml.metric_tensor(circuit)
    >>> weights = np.array([0.1, 0.2, 0.4, 0.5], requires_grad=True)
    >>> mt_fn(weights)
    tensor([[ 0.25  ,  0.    , -0.0497, -0.0497],
            [ 0.    ,  0.2475,  0.0243,  0.0243],
            [-0.0497,  0.0243,  0.0123,  0.0123],
            [-0.0497,  0.0243,  0.0123,  0.0123]], requires_grad=True)

    In order to save cost, one might want to compute only the block-diagonal part of
    the metric tensor, which requires significantly fewer executions of quantum functions
    and does not need an auxiliary wire on the device. This can be done using the
    ``approx`` keyword:

    >>> mt_fn = qml.metric_tensor(circuit, approx="block-diag")
    >>> weights = np.array([0.1, 0.2, 0.4, 0.5], requires_grad=True)
    >>> mt_fn(weights)
    tensor([[0.25  , 0.    , 0.    , 0.    ],
            [0.    , 0.2475, 0.    , 0.    ],
            [0.    , 0.    , 0.0123, 0.0123],
            [0.    , 0.    , 0.0123, 0.0123]], requires_grad=True)

    These blocks are given by parameter groups that belong to groups of commuting gates.

    The tensor can be further restricted to the diagonal via ``approx="diag"``. However,
    this will not save further quantum function evolutions but only classical postprocessing.

    The returned metric tensor is also fully differentiable in all interfaces.
    For example, we can compute the gradient of the Frobenius norm of the metric tensor
    with respect to the QNode ``weights`` :

    >>> norm_fn = lambda x: qml.math.linalg.norm(mt_fn(x), ord="fro")
    >>> grad_fn = qml.grad(norm_fn)
    >>> grad_fn(weights)
    array([-0.0282246 ,  0.01340413,  0.        ,  0.        ])

    .. details::
        :title: Usage Details

        This transform can also be applied to low-level
        :class:`~.QuantumTape` objects. This will result in no implicit quantum
        device evaluation. Instead, the processed tapes, and post-processing
        function, which together define the metric tensor are directly returned:

        >>> params = np.array([1.7, 1.0, 0.5], requires_grad=True)
        >>> with qml.tape.QuantumTape() as tape:
        ...     qml.RX(params[0], wires=0)
        ...     qml.RY(params[1], wires=0)
        ...     qml.CNOT(wires=[0, 1])
        ...     qml.PhaseShift(params[2], wires=1)
        ...     qml.expval(qml.PauliX(0))
        >>> tapes, fn = qml.metric_tensor(tape)
        >>> tapes
        [<QuantumTape: wires=[0, 1], params=0>,
         <QuantumTape: wires=[0, 1], params=1>,
         <QuantumTape: wires=[0, 1], params=3>,
         <QuantumTape: wires=[2, 0], params=1>,
         <QuantumTape: wires=[2, 0, 1], params=2>,
         <QuantumTape: wires=[2, 0, 1], params=2>]

        This can be useful if the underlying circuits representing the metric tensor
        computation need to be analyzed. We clearly can distinguish the first three
        tapes used for the block-diagonal from the last three tapes that use the
        auxiliary wire ``2`` , which was not used by the original tape.

        The output tapes can then be evaluated and post-processed to retrieve
        the metric tensor:

        >>> dev = qml.device("default.qubit", wires=2)
        >>> fn(qml.execute(tapes, dev, None))
        array([[ 0.25      ,  0.        ,  0.42073549],
               [ 0.        ,  0.00415023, -0.26517488],
               [ 0.42073549, -0.26517488,  0.24878844]])

        The first term of the off block-diagonal entries of the full metric tensor are
        computed with Hadamard tests. This first term reads

        .. math ::

            \mathfrak{Re}\left\{\langle \partial_i\psi|\partial_j\psi\rangle\right\}

        and can be computed using an augmented circuit with an additional qubit.
        See for example the appendix of `McArdle et al. (2019) <https://doi.org/10.1038/s41534-019-0187-2>`__
        for details.
        The block-diagonal of the tensor is computed using the covariance matrix approach.

        In addition, we may extract the factors for the second terms
        :math:`\langle \psi|\partial_j\psi\rangle`
        of the *off block-diagonal* tensor from the quantum function output for the covariance matrix!

        This means that in total only the tapes for the first terms of the off block-diagonal
        are required in addition to the circuits for the block diagonal.
    """
    if not tape.trainable_params:
        warnings.warn(
            "Attempted to compute the metric tensor of a tape with no trainable parameters. "
            "If this is unintended, please mark trainable parameters in accordance with the "
            "chosen auto differentiation framework, or via the 'tape.trainable_params' property."
        )
        return [], lambda _: ()

    # pylint: disable=too-many-arguments
    if approx in {"diag", "block-diag"}:
        # Only require covariance matrix based transform
        diag_approx = approx == "diag"
        return _metric_tensor_cov_matrix(tape, diag_approx)[:2]

    if approx is None:
        return _metric_tensor_hadamard(tape, allow_nonunitary, aux_wire, device_wires)

    raise ValueError(
        f"Unknown value {approx} for keyword argument approx. "
        "Valid values are 'diag', 'block-diag' and None."
    )


def _contract_metric_tensor_with_cjac(mt, cjac):
    """Execute the contraction of pre-computed classical Jacobian(s)
    and the metric tensor of a tape in order to obtain the hybrid
    metric tensor of a QNode.

    Args:
        mt (array): Metric tensor of a tape (2-dimensional)
        cjac (array or tuple[array]): The classical Jacobian of a QNode

    Returns:
        array or tuple[array]: Hybrid metric tensor(s) of the QNode.
        The number of metric tensors depends on the number of QNode arguments
        for which the classical Jacobian was computed, the tensor shape(s)
        depend on the shape of these QNode arguments.
    """
    if isinstance(cjac, tuple):
        # Classical processing of multiple arguments is present. Return cjac.T @ mt @ cjac
        # as a tuple of contractions.
        metric_tensors = tuple(
            qml.math.tensordot(c, qml.math.tensordot(mt, c, axes=[[-1], [0]]), axes=[[0], [0]])
            for c in cjac
            if c is not None
        )
        return metric_tensors[0] if len(metric_tensors) == 1 else metric_tensors

    is_square = cjac.shape == (1,) or (cjac.ndim == 2 and cjac.shape[0] == cjac.shape[1])

    if is_square and qml.math.allclose(cjac, qml.numpy.eye(cjac.shape[0])):
        # Classical Jacobian is the identity. No classical processing
        # is present inside the QNode.
        return mt

    mt = qml.math.tensordot(cjac, qml.math.tensordot(mt, cjac, axes=[[-1], [0]]), axes=[[0], [0]])

    return mt


@metric_tensor.custom_qnode_wrapper
def qnode_execution_wrapper(self, qnode, targs, tkwargs):
    """Here, we overwrite the QNode execution wrapper in order
    to take into account that classical processing may be present
    inside the QNode."""
    hybrid = tkwargs.pop("hybrid", True)

    if isinstance(qnode, qml.ExpvalCost):
        if qnode._multiple_devices:  # pylint: disable=protected-access
            warnings.warn(
                "ExpvalCost was instantiated with multiple devices. Only the first device "
                "will be used to evaluate the metric tensor.",
                UserWarning,
            )

        qnode = qnode.qnodes.qnodes[0]

    tkwargs.setdefault("device_wires", qnode.device.wires)
    mt_fn = self.default_qnode_wrapper(qnode, targs, tkwargs)

    def _expand_fn(tape):
        return self.expand_fn(tape, *targs, **tkwargs)

    cjac_fn = qml.transforms.classical_jacobian(qnode, expand_fn=_expand_fn)

    def wrapper(*args, **kwargs):
        if not qml.math.get_trainable_indices(args):
            warnings.warn(
                "Attempted to compute the metric tensor of a QNode with no trainable parameters. "
                "If this is unintended, please add trainable parameters in accordance with the "
                "chosen auto differentiation framework."
            )
            return ()

        try:
            mt = mt_fn(*args, **kwargs)
        except qml.wires.WireError as e:
            if str(e) == "No device wires are unused by the tape.":
                warnings.warn(
                    "The device does not have a wire that is not used by the tape."
                    "\n\nReverting to the block-diagonal approximation. It will often be "
                    "much more efficient to request the block-diagonal approximation directly!"
                )
            else:
                warnings.warn(
                    "An auxiliary wire is not available."
                    "\n\nThis can occur when computing the full metric tensor via the "
                    "Hadamard test, and the device does not provide an "
                    "additional wire or the requested auxiliary wire does not exist "
                    "on the device."
                    "\n\nReverting to the block-diagonal approximation. It will often be "
                    "much more efficient to request the block-diagonal approximation directly!"
                )
            tkwargs["approx"] = "block-diag"
            return self(qnode, *targs, **tkwargs)(*args, **kwargs)

        if not hybrid:
            return mt

        kwargs.pop("shots", False)
        cjac = cjac_fn(*args, **kwargs)

        return _contract_metric_tensor_with_cjac(mt, cjac)

    return wrapper


def _metric_tensor_cov_matrix(tape, diag_approx):
    r"""This is the metric tensor method for the block diagonal, using
    the covariance matrix of the generators of each layer.

    Args:
        tape (pennylane.QNode or .QuantumTape): quantum tape or QNode to find the metric tensor of
        diag_approx (bool): if True, use the diagonal approximation. If ``False`` , a
            block-diagonal approximation of the metric tensor is computed.
    Returns:
        list[pennylane.tape.QuantumTape]: Transformed tapes that compute the probabilities
            required for the covariance matrix
        callable: Post-processing function that computes the covariance matrix from the
            results of the tapes in the first return value
        list[list[.Observable]]: Observables measured in each tape, one inner list
            corresponding to one tape in the first return value
        list[list[float]]: Coefficients to scale the results for each observable, one inner list
            corresponding to one tape in the first return value

    This method assumes the ``generator`` of all parametrized operations with respect to
    which the tensor is computed to be Hermitian. This is the case for unitary single-parameter
    operations.
    """
    # get the circuit graph
    graph = tape.graph

    metric_tensor_tapes = []
    obs_list = []
    coeffs_list = []
    params_list = []

    for queue, curr_ops, param_idx, _ in graph.iterate_parametrized_layers():
        params_list.append(param_idx)
        coeffs_list.append([])
        obs_list.append([])

        # for each operation in the layer, get the generator
        for op in curr_ops:
            obs, s = qml.generator(op)
            obs_list[-1].append(obs)
            coeffs_list[-1].append(s)

        # Create a quantum tape with all operations
        # prior to the parametrized layer, and the rotations
        # to measure in the basis of the parametrized layer generators.
        with qml.queuing.AnnotatedQueue() as layer_q:
            for op in queue:
                qml.apply(op)

            for o in obs_list[-1]:
                o.diagonalizing_gates()

            qml.probs(wires=tape.wires)

        layer_tape = qml.tape.QuantumScript.from_queue(layer_q)
        metric_tensor_tapes.append(layer_tape)

    def processing_fn(probs):
        gs = []

        for prob, obs, coeffs in zip(probs, obs_list, coeffs_list):

            # calculate the covariance matrix of this layer
            scale = qml.math.convert_like(qml.math.outer(coeffs, coeffs), prob)
            scale = qml.math.cast_like(scale, prob)
            g = scale * qml.math.cov_matrix(prob, obs, wires=tape.wires, diag_approx=diag_approx)
            gs.append(g)

        # create the block diagonal metric tensor
        return qml.math.block_diag(gs)

    return metric_tensor_tapes, processing_fn, obs_list, coeffs_list


@functools.lru_cache()
def _get_gen_op(op, allow_nonunitary, aux_wire):
    r"""Get the controlled-generator operation for a given operation.

    Args:
        op (pennylane.operation.Operation): Operation from which to extract the generator
        allow_nonunitary (bool): Whether non-unitary gates are allowed in the circuit
        aux_wire (int or pennylane.wires.Wires): Auxiliary wire on which to control the operation

    Returns
        qml.Operation: Controlled-generator operation of the generator of ``op``, controlled
        on wire ``aux_wire``.

    Raises
        ValueError: If the generator of ``op`` is not known or it is non-unitary while
        ``allow_nonunitary=False``.

    If ``allow_nonunitary=True``, a general :class:`~.pennylane.ControlledQubitUnitary` is returned,
    otherwise only controlled Pauli operations are used. If the operation has a non-unitary
    generator but ``allow_nonunitary=False``, the operation ``op`` should have been decomposed
    before, leading to a ``ValueError``.
    """
    op_to_cgen = {
        qml.RX: qml.CNOT,
        qml.RY: qml.CY,
        qml.RZ: qml.CZ,
        qml.PhaseShift: qml.CZ,  # PhaseShift is the same as RZ up to a global phase
    }

    try:
        cgen = op_to_cgen[op.__class__]
        return cgen(wires=[aux_wire, *op.wires])

    except KeyError as e:
        if allow_nonunitary:
            mat = qml.matrix(qml.generator(op, format="observable"))
            return qml.ControlledQubitUnitary(mat, control_wires=aux_wire, wires=op.wires)

        raise ValueError(
            f"Generator for operation {op} not known and non-unitary operations "
            "deactivated via allow_nonunitary=False."
        ) from e


def _get_first_term_tapes(layer_i, layer_j, allow_nonunitary, aux_wire):
    r"""Obtain the tapes for the first term of all tensor entries
    belonging to an off-diagonal block.

    Args:
        layer_i (list): The first layer of parametrized ops, of the format of
            the layers generated by ``iterate_parametrized_layers``
        layer_j (list): The second layer of parametrized ops
        allow_nonunitary (bool): Whether non-unitary operations are allowed
            in the circuit; passed to ``_get_gen_op``
        aux_wire (object or pennylane.wires.Wires): Auxiliary wire on which to
            control the controlled-generator operations

    Returns:
        list[pennylane.tape.QuantumTape]: Transformed tapes that compute the
            first term of the metric tensor for the off-diagonal block belonging
            to the input layers
        list[tuple[int]]: 2-tuple indices assigning the tapes to metric tensor
            entries
    """

    tapes = []
    ids = []
    # Exclude the backwards cone of layer_i from the backwards cone of layer_j
    ops_between_cgens = [op for op in layer_j.pre_ops if op not in layer_i.pre_ops]

    # Iterate over differentiated operation in first layer
    for diffed_op_i, par_idx_i in zip(layer_i.ops, layer_i.param_inds):
        gen_op_i = _get_gen_op(diffed_op_i, allow_nonunitary, aux_wire)

        # Iterate over differentiated operation in second layer
        # There will be one tape per pair of differentiated operations
        for diffed_op_j, par_idx_j in zip(layer_j.ops, layer_j.param_inds):
            gen_op_j = _get_gen_op(diffed_op_j, allow_nonunitary, aux_wire)

            with qml.queuing.AnnotatedQueue() as q:
                # Initialize auxiliary wire
                qml.Hadamard(wires=aux_wire)
                # Apply backward cone of first layer
                for op in layer_i.pre_ops:
                    qml.apply(op)
                # Controlled-generator operation of first diff'ed op
                qml.apply(gen_op_i)
                # Apply first layer and operations between layers
                for op in ops_between_cgens:
                    qml.apply(op)
                # Controlled-generator operation of second diff'ed op
                qml.apply(gen_op_j)
                # Measure X on auxiliary wire
                qml.expval(qml.PauliX(aux_wire))

            tapes.append(qml.tape.QuantumScript.from_queue(q))
            # Memorize to which metric entry this tape belongs
            ids.append((par_idx_i, par_idx_j))

    return tapes, ids


def _metric_tensor_hadamard(tape, allow_nonunitary, aux_wire, device_wires):
    r"""Generate the quantum tapes that execute the Hadamard tests
    to compute the first term of off block-diagonal metric entries
    and combine them with the covariance matrix-based block-diagonal tapes.

    Args:
        tape (pennylane.QNode or .QuantumTape): quantum tape or QNode to find the metric tensor of
        allow_nonunitary (bool): Whether non-unitary operations are allowed in circuits
            created by the transform. Only relevant if ``approx`` is ``None``
            Should be set to ``True`` if possible to reduce cost.
        aux_wire (int or .wires.Wires): Auxiliary wire to be used for
            Hadamard tests. By default, a suitable wire is inferred from the number
            of used wires in the original circuit.
        device_wires (.wires.Wires): Wires of the device that is going to be used for the
            metric tensor. Facilitates finding a default for ``aux_wire`` if ``aux_wire``
            is ``None`` .

    Returns:
        list[pennylane.tape.QuantumTape]: Tapes to evaluate the metric tensor
        callable: processing function to obtain the metric tensor from the tape results
    """
    # Get tapes and processing function for the block-diagonal metric tensor,
    # as well as the generator observables and generator coefficients for each diff'ed operation
    diag_tapes, diag_proc_fn, obs_list, coeffs = _metric_tensor_cov_matrix(tape, diag_approx=False)

    # Obtain layers of parametrized operations and account for the discrepancy between trainable
    # and non-trainable parameter indices
    graph = tape.graph
    par_idx_to_trainable_idx = {idx: i for i, idx in enumerate(sorted(tape.trainable_params))}
    layers = [
        LayerData(
            layer.pre_ops,
            layer.ops,
            tuple(par_idx_to_trainable_idx[idx] for idx in layer[2]),
            layer.post_ops,
        )
        for layer in graph.iterate_parametrized_layers()
    ]
    if len(layers) <= 1:
        return diag_tapes, diag_proc_fn

    # Get default for aux_wire
    aux_wire = _get_aux_wire(aux_wire, tape, device_wires)

    # Get all tapes for the first term of the metric tensor and memorize which
    # entry they belong to
    first_term_tapes = []
    ids = []
    block_sizes = []
    for idx_i, layer_i in enumerate(layers):
        block_sizes.append(len(layer_i.param_inds))

        for layer_j in layers[idx_i + 1 :]:
            _tapes, _ids = _get_first_term_tapes(layer_i, layer_j, allow_nonunitary, aux_wire)
            first_term_tapes.extend(_tapes)
            ids.extend(_ids)

    # Combine block-diagonal and off block-diagonal tapes
    tapes = diag_tapes + first_term_tapes
    # prepare off block-diagonal mask
    mask = 1 - qml.math.block_diag([qml.math.ones((bsize, bsize)) for bsize in block_sizes])
    # Required for slicing in processing_fn
    num_diag_tapes = len(diag_tapes)

    def processing_fn(results):
        """Postprocessing function for the full metric tensor."""
        nonlocal mask
        # Split results
        diag_res, off_diag_res = results[:num_diag_tapes], results[num_diag_tapes:]
        # Get full block-diagonal tensor
        diag_mt = diag_proc_fn(diag_res)

        if qml.active_return():
            # the off diag tapes only have a single expval measurement
            off_diag_res = [qml.math.expand_dims(res, 0) for res in off_diag_res]

        # Prepare the mask to match the used interface
        mask = qml.math.convert_like(mask, diag_mt)

        # Initialize off block-diagonal tensor using the stored ids
        first_term = qml.math.zeros_like(diag_mt)
        if ids:
            off_diag_res = qml.math.stack(off_diag_res, 1)[0]
            inv_ids = [_id[::-1] for _id in ids]
            first_term = qml.math.scatter_element_add(first_term, list(zip(*ids)), off_diag_res)
            first_term = qml.math.scatter_element_add(first_term, list(zip(*inv_ids)), off_diag_res)

        # Second terms of off block-diagonal metric tensor
        expvals = qml.math.zeros_like(first_term[0])
        idx = 0
        for prob, obs in zip(diag_res, obs_list):
            for o in obs:
                l = qml.math.cast(o.eigvals(), dtype=np.float64)
                w = tape.wires.indices(o.wires)
                p = qml.math.marginal_prob(prob, w)
                expvals = qml.math.scatter_element_add(expvals, (idx,), qml.math.dot(l, p))
                idx += 1

        # Construct <\partial_i\psi|\psi><\psi|\partial_j\psi> and mask it
        second_term = qml.math.tensordot(expvals, expvals, axes=0) * mask

        # Subtract second term from first term
        off_diag_mt = first_term - second_term

        # Rescale first and second term
        _coeffs = qml.math.hstack(coeffs)
        scale = qml.math.convert_like(qml.math.tensordot(_coeffs, _coeffs, axes=0), results[0])
        off_diag_mt = scale * off_diag_mt

        # Combine block-diagonal and off block-diagonal
        mt = off_diag_mt + diag_mt

        return mt

    return tapes, processing_fn


def _get_aux_wire(aux_wire, tape, device_wires):
    r"""Determine an unused wire to be used as auxiliary wire for Hadamard tests.

    Args:
        aux_wire (object): Input auxiliary wire. Returned unmodified if not ``None``
        tape (pennylane.tape.QuantumTape): Tape to infer the wire for
        device_wires (.wires.Wires): Wires of the device that is going to be used for the
            metric tensor. Facilitates finding a default for ``aux_wire`` if ``aux_wire``
            is ``None`` .

    Returns:
        object: The auxiliary wire to be used. Equals ``aux_wire`` if it was not ``None`` ,
        and an often reasonable choice else.
    """
    if aux_wire is not None:
        if device_wires is None or aux_wire in device_wires:
            return aux_wire
        raise qml.wires.WireError("The requested aux_wire does not exist on the used device.")

    if device_wires is not None:
        unused_wires = qml.wires.Wires(device_wires.toset().difference(tape.wires.toset()))
        if not unused_wires:
            raise qml.wires.WireError("No device wires are unused by the tape.")
        return unused_wires[0]

    _wires = tape.wires
    for _aux in range(tape.num_wires):
        if _aux not in _wires:
            return _aux

    return tape.num_wires



# Copyright 2018-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for computing Autograd gradients of quantum functions.
"""

import pytest
import autograd
import pennylane.numpy as anp  # only to be used inside classical computational nodes
import numpy as np

import pennylane as qml
from gate_data import Rotx as Rx, Roty as Ry, Rotz as Rz


alpha = 0.5  # displacement in tests
hbar = 2
mag_alphas = np.linspace(0, 1.5, 5)
thetas = np.linspace(-2 * np.pi, 2 * np.pi, 8)
sqz_vals = np.linspace(0.0, 1.0, 5)


class PolyN(qml.ops.PolyXP):
    "Mimics NumberOperator using the arbitrary 2nd order observable interface. Results should be identical."

    def __init__(self, wires):
        hbar = 2
        q = np.diag([-0.5, 0.5 / hbar, 0.5 / hbar])
        super().__init__(q, wires=wires)
        self.name = "PolyXP"


@pytest.fixture(scope="module")
def gaussian_dev():
    return qml.device("default.gaussian", wires=2)


@pytest.fixture(scope="module")
def grad_fn_R(gaussian_dev):
    @qml.qnode(gaussian_dev)
    def circuit(y):
        qml.Displacement(alpha, 0.0, wires=[0])
        qml.Rotation(y, wires=[0])
        return qml.expval(qml.X(0))

    return autograd.grad(circuit)


@pytest.fixture(scope="module")
def grad_fn_BS(gaussian_dev):
    @qml.qnode(gaussian_dev)
    def circuit(y):
        qml.Displacement(alpha, 0.0, wires=[0])
        qml.Beamsplitter(y, 0, wires=[0, 1])
        return qml.expval(qml.X(0))

    return autograd.grad(circuit)


@pytest.fixture(scope="module")
def grad_fn_D(gaussian_dev):
    @qml.qnode(gaussian_dev)
    def circuit(r, phi):
        qml.Displacement(r, phi, wires=[0])
        return qml.expval(qml.X(0))

    return autograd.grad(circuit)


@pytest.fixture(scope="module")
def grad_fn_S(gaussian_dev):
    @qml.qnode(gaussian_dev)
    def circuit(y):
        qml.Displacement(alpha, 0.0, wires=[0])
        qml.Squeezing(y, 0.0, wires=[0])
        return qml.expval(qml.X(0))

    return autograd.grad(circuit)


@pytest.fixture(scope="module")
def grad_fn_S_Fock(gaussian_dev):
    @qml.qnode(gaussian_dev)
    def circuit(y):
        qml.Squeezing(y, 0.0, wires=[0])
        return qml.expval(qml.FockStateProjector(np.array([2, 0]), wires=[0, 1]))

    return autograd.grad(circuit)


class TestCVGradient:
    """Tests of the automatic gradient method for CV circuits."""

    @pytest.mark.parametrize("theta", thetas)
    def test_rotation_gradient(self, theta, grad_fn_R, tol):
        "Tests that the automatic gradient of a phase space rotation is correct."

        autograd_val = grad_fn_R(theta)
        # qfunc evalutes to hbar * alpha * cos(theta)
        manualgrad_val = -hbar * alpha * np.sin(theta)
        assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    @pytest.mark.parametrize("theta", thetas)
    def test_beamsplitter_gradient(self, theta, grad_fn_BS, tol):
        "Tests that the automatic gradient of a beamsplitter is correct."

        autograd_val = grad_fn_BS(theta)
        # qfunc evalutes to hbar * alpha * cos(theta)
        manualgrad_val = -hbar * alpha * np.sin(theta)
        assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    @pytest.mark.parametrize("mag", mag_alphas)
    @pytest.mark.parametrize("theta", thetas)
    def test_displacement_gradient(self, mag, theta, grad_fn_D, tol):
        "Tests that the automatic gradient of a phase space displacement is correct."

        # alpha = mag * np.exp(1j * theta)
        autograd_val = grad_fn_D(mag, theta)
        # qfunc evalutes to hbar * Re(alpha)
        manualgrad_val = hbar * np.cos(theta)
        assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    @pytest.mark.parametrize("r", sqz_vals)
    def test_squeeze_gradient(self, r, grad_fn_S, tol):
        "Tests that the automatic gradient of a phase space squeezing is correct."

        autograd_val = grad_fn_S(r)
        # qfunc evaluates to -exp(-r) * hbar * Re(alpha)
        manualgrad_val = -np.exp(-r) * hbar * alpha
        assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    @pytest.mark.parametrize("r", sqz_vals[1:])  # formula is not valid for r=0
    def test_number_state_gradient(self, r, grad_fn_S_Fock, tol):
        "Tests that the automatic gradient of a squeezed state with number state expectation is correct."

        # (d/dr) |<2|S(r)>|^2 = 0.5 tanh(r)^3 (2 csch(r)^2 - 1) sech(r)
        autograd_val = grad_fn_S_Fock(r)
        manualgrad_val = 0.5 * np.tanh(r) ** 3 * (2 / (np.sinh(r) ** 2) - 1) / np.cosh(r)
        assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    @pytest.mark.autograd
    @pytest.mark.parametrize("O", [qml.ops.X, qml.ops.NumberOperator])
    @pytest.mark.parametrize(
        "make_gate",
        [lambda x: qml.Rotation(x, wires=0), lambda x: qml.ControlledPhase(x, wires=[0, 1])],
    )
    def test_cv_gradients_gaussian_circuit(self, make_gate, O, gaussian_dev, tol):
        """Tests that the gradients of circuits of gaussian gates match
        between the finite difference and analytic methods."""

        tol = 1e-5
        par = anp.array(0.4, requires_grad=True)

        def circuit(x):
            qml.Displacement(0.5, 0, wires=0)
            make_gate(x)
            qml.Beamsplitter(1.3, -2.3, wires=[0, 1])
            qml.Displacement(-0.5, 0.1, wires=0)
            qml.Squeezing(0.5, -1.5, wires=0)
            qml.Rotation(-1.1, wires=0)
            return qml.expval(O(wires=0))

        q = qml.QNode(circuit, gaussian_dev)
        val = q(par)

        grad_F = qml.gradients.finite_diff(q)(par)
        grad_A2 = qml.gradients.param_shift_cv(q, dev=gaussian_dev, force_order2=True)(par)
        if O.ev_order == 1:
            grad_A = qml.gradients.param_shift_cv(q, dev=gaussian_dev)(par)
            # the different methods agree
            assert qml.math.shape(grad_A) == qml.math.shape(grad_F) == ()
            assert np.allclose(grad_A, grad_F, atol=tol)

        # the different methods agree
        assert qml.math.shape(grad_A2) == qml.math.shape(grad_F) == ()
        assert np.allclose(grad_A2, grad_F, atol=tol)

    @pytest.mark.autograd
    def test_cv_gradients_multiple_gate_parameters(self, gaussian_dev, tol):
        "Tests that gates with multiple free parameters yield correct gradients."
        par = anp.array([0.4, -0.3, -0.7, 0.2], requires_grad=True)

        def qf(r0, phi0, r1, phi1):
            qml.Squeezing(r0, phi0, wires=[0])
            qml.Squeezing(r1, phi1, wires=[0])
            return qml.expval(qml.NumberOperator(0))

        q = qml.QNode(qf, gaussian_dev)
        q(*par)
        grad_F = qml.gradients.finite_diff(q)(*par)
        grad_A = qml.gradients.param_shift_cv(q, dev=gaussian_dev)(*par)
        grad_A2 = qml.gradients.param_shift_cv(q, dev=gaussian_dev, force_order2=True)(*par)

        # the different methods agree
        assert qml.math.allclose(grad_A, grad_F, atol=tol, rtol=0)
        assert qml.math.allclose(grad_A2, grad_F, atol=tol, rtol=0)

        # check against the known analytic formula
        r0, phi0, r1, phi1 = par
        dn = np.zeros([4])
        dn[0] = np.cosh(2 * r1) * np.sinh(2 * r0) + np.cos(phi0 - phi1) * np.cosh(2 * r0) * np.sinh(
            2 * r1
        )
        dn[1] = -0.5 * np.sin(phi0 - phi1) * np.sinh(2 * r0) * np.sinh(2 * r1)
        dn[2] = np.cos(phi0 - phi1) * np.cosh(2 * r1) * np.sinh(2 * r0) + np.cosh(2 * r0) * np.sinh(
            2 * r1
        )
        dn[3] = 0.5 * np.sin(phi0 - phi1) * np.sinh(2 * r0) * np.sinh(2 * r1)

        assert all(qml.math.isclose(dn[i], grad_F[i], atol=tol, rtol=0) for i in range(4))

    @pytest.mark.autograd
    def test_cv_gradients_repeated_gate_parameters(self, gaussian_dev, tol):
        "Tests that repeated use of a free parameter in a multi-parameter gate yield correct gradients."
        par = anp.array([0.2, 0.3], requires_grad=True)

        def qf(x, y):
            qml.Displacement(x, 0, wires=[0])
            qml.Squeezing(y, -1.3 * y, wires=[0])
            return qml.expval(qml.X(0))

        q = qml.QNode(qf, gaussian_dev)
        q(*par)
        grad_F = qml.gradients.finite_diff(q)(*par)
        grad_A = qml.gradients.param_shift_cv(q, dev=gaussian_dev)(*par)
        grad_A2 = qml.gradients.param_shift_cv(q, dev=gaussian_dev, force_order2=True)(*par)

        # the different methods agree
        assert qml.math.allclose(grad_A, grad_F, atol=tol, rtol=0)
        assert qml.math.allclose(grad_A2, grad_F, atol=tol, rtol=0)

    @pytest.mark.autograd
    def test_cv_gradients_parameters_inside_array(self, gaussian_dev, tol):
        "Tests that free parameters inside an array passed to an Operation yield correct gradients."
        par = anp.array([0.4, 1.3], requires_grad=True)

        def qf(x, y):
            qml.Displacement(0.5, 0, wires=[0])
            qml.Squeezing(x, 0, wires=[0])
            M = np.zeros((5, 5))
            M[1, 1] = y
            M[1, 2] = 1.0
            M[2, 1] = 1.0
            return qml.expval(qml.PolyXP(M, [0, 1]))

        q = qml.QNode(qf, gaussian_dev)
        q(*par)
        grad_F = qml.gradients.finite_diff(q, hybrid=False)(*par)
        grad_A = qml.gradients.param_shift_cv(q, dev=gaussian_dev, hybrid=False)(*par)
        grad_A2 = qml.gradients.param_shift_cv(
            q, dev=gaussian_dev, force_order2=True, hybrid=False
        )(*par)

        # the different methods agree
        assert grad_A2 == pytest.approx(grad_F, abs=tol)

    @pytest.mark.autograd
    def test_cv_gradient_fanout(self, gaussian_dev, tol):
        "Tests that qnodes can compute the correct gradient when the same parameter is used in multiple gates."
        par = anp.array([0.5, 1.3], requires_grad=True)

        def circuit(x, y):
            qml.Displacement(x, 0, wires=[0])
            qml.Rotation(y, wires=[0])
            qml.Displacement(0, x, wires=[0])
            return qml.expval(qml.X(0))

        q = qml.QNode(circuit, gaussian_dev)
        q(*par)
        grad_F = qml.gradients.finite_diff(q)(*par)
        grad_A = qml.gradients.param_shift_cv(q, dev=gaussian_dev)(*par)
        grad_A2 = qml.gradients.param_shift_cv(q, dev=gaussian_dev, force_order2=True)(*par)

        # the different methods agree
        assert qml.math.allclose(grad_A, grad_F, atol=tol, rtol=0)
        assert qml.math.allclose(grad_A2, grad_F, atol=tol, rtol=0)

    @pytest.mark.autograd
    def test_CVOperation_with_heisenberg_and_no_params(self, gaussian_dev, tol):
        """An integration test for InterferometerUnitary, a gate that supports analytic
        differentiation if succeeding the gate to be differentiated, but cannot be
        differentiated itself.

        This ensures that, assuming the _heisenberg_rep is defined, the quantum
        gradient analytic method can still be used, and returns the correct result.
        """

        x = anp.array(0.5, requires_grad=True)

        U = np.array(
            [
                [0.51310276 + 0.81702166j, 0.13649626 + 0.22487759j],
                [0.26300233 + 0.00556194j, -0.96414101 - 0.03508489j],
            ]
        )

        def circuit(x):
            qml.Displacement(x, 0, wires=0)
            qml.InterferometerUnitary(U, wires=[0, 1])
            return qml.expval(qml.X(0))

        qnode = qml.QNode(circuit, gaussian_dev)
        qnode(x)
        grad_F = qml.gradients.finite_diff(qnode)(x)
        grad_A = qml.gradients.param_shift_cv(qnode, dev=gaussian_dev)(x)
        grad_A2 = qml.gradients.param_shift_cv(qnode, dev=gaussian_dev, force_order2=True)(x)

        # the different methods agree
        assert qml.math.shape(grad_A) == qml.math.shape(grad_A2) == qml.math.shape(grad_F) == ()
        assert np.allclose(grad_A, grad_F, atol=tol)
        assert np.allclose(grad_A2, grad_F, atol=tol)


class TestQubitGradient:
    """Tests of the automatic gradient method for qubit gates."""

    def test_RX_gradient(self, qubit_device_1_wire, tol):
        "Tests that the automatic gradient of a Pauli X-rotation is correct."

        @qml.qnode(qubit_device_1_wire)
        def circuit(x):
            qml.RX(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        grad_fn = autograd.grad(circuit)

        for theta in thetas:
            autograd_val = grad_fn(theta)
            manualgrad_val = (circuit(theta + np.pi / 2) - circuit(theta - np.pi / 2)) / 2
            assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    def test_RY_gradient(self, qubit_device_1_wire, tol):
        "Tests that the automatic gradient of a Pauli Y-rotation is correct."

        @qml.qnode(qubit_device_1_wire)
        def circuit(x):
            qml.RY(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        grad_fn = autograd.grad(circuit)

        for theta in thetas:
            autograd_val = grad_fn(theta)
            manualgrad_val = (circuit(theta + np.pi / 2) - circuit(theta - np.pi / 2)) / 2
            assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    def test_RZ_gradient(self, qubit_device_1_wire, tol):
        "Tests that the automatic gradient of a Pauli Z-rotation is correct."

        @qml.qnode(qubit_device_1_wire)
        def circuit(x):
            qml.RZ(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        grad_fn = autograd.grad(circuit)

        for theta in thetas:
            autograd_val = grad_fn(theta)
            manualgrad_val = (circuit(theta + np.pi / 2) - circuit(theta - np.pi / 2)) / 2
            assert autograd_val == pytest.approx(manualgrad_val, abs=tol)

    def test_Rot(self, qubit_device_1_wire, tol):
        "Tests that the automatic gradient of a arbitrary Euler-angle-parameterized gate is correct."

        @qml.qnode(qubit_device_1_wire)
        def circuit(x, y, z):
            qml.Rot(x, y, z, wires=[0])
            return qml.expval(qml.PauliZ(0))

        grad_fn = autograd.grad(circuit, argnum=[0, 1, 2])

        eye = np.eye(3)
        for theta in thetas:
            angle_inputs = np.array([theta, theta**3, np.sqrt(2) * theta])
            autograd_val = grad_fn(*angle_inputs)
            for idx in range(3):
                onehot_idx = eye[idx]
                param1 = angle_inputs + np.pi / 2 * onehot_idx
                param2 = angle_inputs - np.pi / 2 * onehot_idx
                manualgrad_val = (circuit(*param1) - circuit(*param2)) / 2
                assert autograd_val[idx] == pytest.approx(manualgrad_val, abs=tol)

    def test_U2(self, tol):
        """Tests that the gradient of an arbitrary U2 gate is correct"""
        dev = qml.device("default.qubit", wires=1)

        @qml.qnode(dev)
        def circuit(x, y):
            qml.QubitStateVector(1j * np.array([1, -1]) / np.sqrt(2), wires=[0])
            qml.U2(x, y, wires=[0])
            return qml.expval(qml.PauliX(0))

        phi = -0.234
        lam = 0.654

        res = circuit(phi, lam)
        expected = np.sin(lam) * np.sin(phi)
        assert np.allclose(res, expected, atol=tol, rtol=0)

        grad_fn = autograd.grad(circuit, argnum=[0, 1])
        res = grad_fn(phi, lam)
        expected = np.array([np.sin(lam) * np.cos(phi), np.cos(lam) * np.sin(phi)])
        assert np.allclose(res, expected, atol=tol, rtol=0)

    def test_U3(self, tol):
        """Tests that the gradient of an arbitrary U3 gate is correct"""
        dev = qml.device("default.qubit", wires=1)

        @qml.qnode(dev)
        def circuit(x, y, z):
            qml.QubitStateVector(1j * np.array([1, -1]) / np.sqrt(2), wires=[0])
            qml.U3(x, y, z, wires=[0])
            return qml.expval(qml.PauliX(0))

        theta = 0.543
        phi = -0.234
        lam = 0.654

        res = circuit(theta, phi, lam)
        expected = np.sin(lam) * np.sin(phi) - np.cos(theta) * np.cos(lam) * np.cos(phi)
        assert np.allclose(res, expected, atol=tol, rtol=0)

        grad_fn = autograd.grad(circuit, argnum=[0, 1, 2])
        res = grad_fn(theta, phi, lam)
        expected = np.array(
            [
                np.sin(theta) * np.cos(lam) * np.cos(phi),
                np.cos(theta) * np.cos(lam) * np.sin(phi) + np.sin(lam) * np.cos(phi),
                np.cos(theta) * np.sin(lam) * np.cos(phi) + np.cos(lam) * np.sin(phi),
            ]
        )
        assert np.allclose(res, expected, atol=tol, rtol=0)

    @pytest.mark.autograd
    def test_qfunc_gradients(self, qubit_device_2_wires, tol):
        "Tests that the various ways of computing the gradient of a qfunc all agree."

        def circuit(x, y, z):
            qml.RX(x, wires=[0])
            qml.CNOT(wires=[0, 1])
            qml.RY(-1.6, wires=[0])
            qml.RY(y, wires=[1])
            qml.CNOT(wires=[1, 0])
            qml.RX(z, wires=[0])
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0))

        qnode = qml.QNode(circuit, qubit_device_2_wires, diff_method="parameter-shift")
        params = anp.array([0.1, -1.6, np.pi / 5], requires_grad=True)

        h = 1e-4 if qubit_device_2_wires.R_DTYPE is np.float32 else 1e-7

        # manual gradients
        grad_fd1 = qml.gradients.finite_diff(qnode, h=h, approx_order=1)(*params)
        grad_fd2 = qml.gradients.finite_diff(qnode, h=h, approx_order=2)(*params)
        grad_angle = qml.gradients.param_shift(qnode)(*params)

        # automatic gradient
        grad_fn = qml.grad(qnode)
        grad_auto = grad_fn(*params)

        # gradients computed with different methods must agree
        assert qml.math.allclose(grad_fd1, grad_fd2, atol=tol, rtol=0)
        assert qml.math.allclose(grad_fd1, grad_angle, atol=tol, rtol=0)
        assert qml.math.allclose(grad_fd1, grad_auto, atol=tol, rtol=0)

    @pytest.mark.autograd
    def test_hybrid_gradients(self, qubit_device_2_wires, tol):
        "Tests that the various ways of computing the gradient of a hybrid computation all agree."

        # input data is the first parameter
        def classifier_circuit(in_data, x):
            qml.RX(in_data, wires=[0])
            qml.CNOT(wires=[0, 1])
            qml.RY(-1.6, wires=[0])
            qml.RY(in_data, wires=[1])
            qml.CNOT(wires=[1, 0])
            qml.RX(x, wires=[0])
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0))

        classifier = qml.QNode(
            classifier_circuit, qubit_device_2_wires, diff_method="parameter-shift"
        )

        param = anp.array(-0.1259, requires_grad=True)
        in_data = qml.numpy.array([-0.1, -0.88, np.exp(0.5)], requires_grad=False)
        out_data = np.array([1.5, np.pi / 3, 0.0])

        def error(p):
            "Total square error of classifier predictions."
            ret = 0
            for d_in, d_out in zip(in_data, out_data):
                square_diff = (classifier(d_in, p) - d_out) ** 2
                ret = ret + square_diff
            return ret

        def d_error(p, grad_method):
            "Gradient of error, computed manually."
            ret = 0
            for d_in, d_out in zip(in_data, out_data):
                args = (d_in, p)
                diff = classifier(*args) - d_out
                ret = ret + 2 * diff * grad_method(classifier)(d_in, p)
            return ret

        y0 = error(param)
        grad = autograd.grad(error)
        grad_auto = grad(param)

        h = 1e-3 if qubit_device_2_wires.R_DTYPE is np.float32 else 1e-7

        grad_fd1 = d_error(param, lambda x: qml.gradients.finite_diff(x, h=h))
        grad_angle = d_error(param, qml.gradients.param_shift)

        # gradients computed with different methods must agree
        assert grad_fd1 == pytest.approx(grad_angle, abs=tol)
        assert grad_fd1 == pytest.approx(grad_auto, abs=tol)
        assert grad_angle == pytest.approx(grad_auto, abs=tol)

    @pytest.mark.autograd
    def test_hybrid_gradients_autograd_numpy(self, qubit_device_2_wires):
        "Test the gradient of a hybrid computation requiring autograd.numpy functions."

        def circuit(x, y):
            "Quantum node."
            qml.RX(x, wires=[0])
            qml.CNOT(wires=[0, 1])
            qml.RY(y, wires=[0])
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))

        quantum = qml.QNode(circuit, qubit_device_2_wires, diff_method="parameter-shift")

        def classical(p):
            "Classical node, requires autograd.numpy functions."
            return anp.exp(anp.sum(quantum(p[0], anp.log(p[1]))))

        def d_classical(a, b, grad_method):
            "Gradient of classical computed symbolically, can use normal numpy functions."
            val = classical((a, b))
            J = grad_method(quantum)(a, np.log(b))
            return val * np.array([J[0][0] + J[0][1], (J[1][0] + J[1][1]) / b])

        param = anp.array([-0.1259, 1.53], requires_grad=True)
        y0 = classical(param)
        grad_classical = autograd.jacobian(classical)
        grad_auto = grad_classical(param)

        h = 5e-4 if qubit_device_2_wires.R_DTYPE is np.float32 else 1e-7
        atol = 2e-3 if qubit_device_2_wires.R_DTYPE is np.float32 else 1e-5

        grad_fd1 = d_classical(*param, lambda x: qml.gradients.finite_diff(x, h=h))
        grad_angle = d_classical(*param, qml.gradients.param_shift)

        # gradients computed with different methods must agree
        assert qml.math.allclose(grad_fd1, grad_angle, atol=atol, rtol=0)
        assert qml.math.allclose(grad_fd1, grad_auto, atol=atol, rtol=0)
        assert qml.math.allclose(grad_angle, grad_auto, atol=atol, rtol=0)

    def test_qnode_gradient_fanout(self, qubit_device_1_wire, tol):
        "Tests that the correct gradient is computed for qnodes which use the same parameter in multiple gates."

        def expZ(state):
            return np.abs(state[0]) ** 2 - np.abs(state[1]) ** 2

        extra_param = 0.31

        def circuit(reused_param, other_param):
            qml.RX(extra_param, wires=[0])
            qml.RY(reused_param, wires=[0])
            qml.RZ(other_param, wires=[0])
            qml.RX(reused_param, wires=[0])
            return qml.expval(qml.PauliZ(0))

        f = qml.QNode(circuit, qubit_device_1_wire)
        zero_state = np.array([1.0, 0.0])

        for reused_p in thetas:
            reused_p = reused_p**3 / 19
            for other_p in thetas:
                other_p = other_p**2 / 11

                # autograd gradient
                grad = autograd.grad(f)
                grad_eval = grad(reused_p, other_p)

                # manual gradient
                grad_true0 = (
                    expZ(
                        Rx(reused_p)
                        @ Rz(other_p)
                        @ Ry(reused_p + np.pi / 2)
                        @ Rx(extra_param)
                        @ zero_state
                    )
                    - expZ(
                        Rx(reused_p)
                        @ Rz(other_p)
                        @ Ry(reused_p - np.pi / 2)
                        @ Rx(extra_param)
                        @ zero_state
                    )
                ) / 2
                grad_true1 = (
                    expZ(
                        Rx(reused_p + np.pi / 2)
                        @ Rz(other_p)
                        @ Ry(reused_p)
                        @ Rx(extra_param)
                        @ zero_state
                    )
                    - expZ(
                        Rx(reused_p - np.pi / 2)
                        @ Rz(other_p)
                        @ Ry(reused_p)
                        @ Rx(extra_param)
                        @ zero_state
                    )
                ) / 2
                grad_true = grad_true0 + grad_true1  # product rule

                assert grad_eval == pytest.approx(grad_true, abs=tol)

    @pytest.mark.autograd
    def test_autograd_trainable_no_warn_grad(self, recwarn):
        """Test that no warning is raised if positional arguments are marked as
        trainable using the requires_grad attribute."""
        dev = qml.device("default.qubit", wires=5)

        @qml.qnode(dev)
        def test(x):
            qml.RZ(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        par = anp.array(0.3, requires_grad=True)
        qml.grad(test)(par)
        assert len(recwarn) == 0

    @pytest.mark.autograd
    def test_autograd_trainable_argnum_no_warn_grad(self, recwarn):
        """Test that no warning is raised if positional arguments are marked as
        trainable using the argnum argument."""
        dev = qml.device("default.qubit", wires=5)

        @qml.qnode(dev)
        def test(x):
            qml.RZ(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        par = np.array(0.3)
        qml.grad(test, argnum=0)(par)
        assert len(recwarn) == 0

    @pytest.mark.autograd
    def test_autograd_trainable_no_warn_jacobian(self, recwarn):
        """Test that no warning is raised if positional arguments are marked as
        trainable using the requires_grad attribute."""
        dev = qml.device("default.qubit", wires=5)

        @qml.qnode(dev)
        def test(x):
            qml.RZ(x, wires=[0])
            return qml.probs(wires=[0])

        par = anp.array(0.3, requires_grad=True)
        qml.jacobian(test)(par)
        assert len(recwarn) == 0

    @pytest.mark.autograd
    def test_autograd_trainable_argnum_no_warn_jacobian(self, recwarn):
        """Test that no warning is raised if positional arguments are marked as
        trainable using the argnum argument."""
        dev = qml.device("default.qubit", wires=5)

        @qml.qnode(dev)
        def test(x):
            qml.RZ(x, wires=[0])
            return qml.probs(wires=[0])

        par = np.array(0.3)
        qml.jacobian(test, argnum=0)(par)
        assert len(recwarn) == 0


class TestFourTermParameterShifts:
    """Tests for quantum gradients that require a 4-term shift formula"""

    @pytest.mark.autograd
    @pytest.mark.parametrize("G", [qml.CRX, qml.CRY, qml.CRZ])
    def test_controlled_rotation_gradient(self, G, tol):
        """Test gradient of controlled RX gate"""
        dev = qml.device("default.qubit", wires=2)
        b = anp.array(0.123, requires_grad=True)

        @qml.qnode(dev, diff_method="parameter-shift")
        def circuit(b):
            qml.QubitStateVector(np.array([1.0, -1.0]) / np.sqrt(2), wires=0)
            G(b, wires=[0, 1])
            return qml.expval(qml.PauliX(0))

        res = circuit(b)
        assert np.allclose(res, -np.cos(b / 2), atol=tol, rtol=0)

        grad = qml.grad(circuit)(b)
        expected = np.sin(b / 2) / 2
        assert np.allclose(grad, expected, atol=tol, rtol=0)

    @pytest.mark.autograd
    @pytest.mark.parametrize("theta", np.linspace(-2 * np.pi, np.pi, 7))
    def test_CRot_gradient(self, theta, tol):
        """Tests that the automatic gradient of a arbitrary controlled Euler-angle-parameterized
        gate is correct."""
        dev = qml.device("default.qubit", wires=2)
        a, b, c = anp.array([theta, theta**3, np.sqrt(2) * theta], requires_grad=True)

        @qml.qnode(dev, diff_method="parameter-shift")
        def circuit(a, b, c):
            qml.QubitStateVector(np.array([1.0, -1.0]) / np.sqrt(2), wires=0)
            qml.CRot(a, b, c, wires=[0, 1])
            return qml.expval(qml.PauliX(0))

        res = circuit(a, b, c)
        expected = -np.cos(b / 2) * np.cos(0.5 * (a + c))
        assert np.allclose(res, expected, atol=tol, rtol=0)

        grad = qml.grad(circuit)(a, b, c)
        expected = np.array(
            [
                [
                    0.5 * np.cos(b / 2) * np.sin(0.5 * (a + c)),
                    0.5 * np.sin(b / 2) * np.cos(0.5 * (a + c)),
                    0.5 * np.cos(b / 2) * np.sin(0.5 * (a + c)),
                ]
            ]
        )
        assert np.allclose(grad, expected, atol=tol, rtol=0)


@pytest.mark.parametrize("diff_method", ["backprop", "parameter-shift", "finite-diff"])
@pytest.mark.parametrize("argnum", [0, 2, [1], [0, 2], [0, 1, 2]])
class TestArgnum:
    """Test various argnum scenarios for qml.grad and qml.jacobian and compare output
    to equivalent requires_grad configuration of the inputs."""

    @pytest.mark.autograd
    @pytest.mark.parametrize(
        "input",
        [
            ((0.1, 0.2), (0.1, 0.2)),
            [[0.1, 0.2], [0.1, 0.2]],
            np.array([[0.1, 0.2], [0.1, 0.2]]),
            qml.numpy.tensor([[0.1, 0.2], [0.1, 0.2]]),
        ],
    )
    def test_grad(self, diff_method, input, argnum):
        """Test qml.grad with various argnums"""
        dev = qml.device("default.qubit", wires=2)

        @qml.qnode(dev, diff_method=diff_method)
        def circuit(x, y, z):
            qml.RX(z[0][1], wires=0)
            qml.RZ(x[0][1], wires=0)
            qml.RY(y[0][1], wires=0)
            return qml.expval(qml.PauliZ(0))

        result = qml.grad(circuit, argnum=argnum)(input, input, input)

        params = [input, input, input]
        for i in range(3):
            if i in ([argnum] if isinstance(argnum, int) else argnum):
                params[i] = qml.numpy.tensor(params[i], requires_grad=True)
            else:
                params[i] = qml.numpy.tensor(params[i], requires_grad=False)

        expected = qml.grad(circuit)(*params)

        assert np.allclose(result, expected)

    @pytest.mark.autograd
    @pytest.mark.parametrize(
        "input",
        [np.array([[0.1, 0.2], [0.1, 0.2]]), qml.numpy.tensor([[0.1, 0.2], [0.1, 0.2]])],
    )
    def test_jacobian(self, diff_method, input, argnum):
        """Test qml.jacobian with various argnums (no support for lists/tuples)"""
        dev = qml.device("default.qubit", wires=2)

        @qml.qnode(dev, diff_method=diff_method)
        def circuit(x, y, z):
            qml.RX(z[0][1], wires=0)
            qml.RZ(x[0][1], wires=0)
            qml.RY(y[0][1], wires=0)
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))

        result = qml.jacobian(circuit, argnum=argnum)(input, input, input)

        params = [input, input, input]
        for i in range(3):
            if i in ([argnum] if isinstance(argnum, int) else argnum):
                params[i] = qml.numpy.tensor(params[i], requires_grad=True)
            else:
                params[i] = qml.numpy.tensor(params[i], requires_grad=False)

        expected = qml.jacobian(circuit)(*params)

        assert np.allclose(result, expected)



# Copyright 2022 Xanadu Quantum Technologies Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Function cut_circuit_mc for cutting a quantum circuit into smaller circuit fragments using a
    Monte Carlo method, at its auxillary functions"""

import inspect
from functools import partial
from typing import Callable, List, Optional, Sequence, Tuple, Union

from networkx import MultiDiGraph

import pennylane as qml
from pennylane import numpy as np
from pennylane.measurements import SampleMP
from pennylane.queuing import AnnotatedQueue
from pennylane.tape import QuantumScript, QuantumTape
from pennylane.transforms.batch_transform import batch_transform
from pennylane.wires import Wires

from .cutstrategy import CutStrategy
from .kahypar import kahypar_cut
from .processing import qcut_processing_fn_mc, qcut_processing_fn_sample
from .qcut import (
    MeasureNode,
    PrepareNode,
    _prep_iminus_state,
    _prep_iplus_state,
    _prep_minus_state,
    _prep_one_state,
    _prep_plus_state,
    _prep_zero_state,
)
from .tapes import _qcut_expand_fn, graph_to_tape, tape_to_graph
from .utils import find_and_place_cuts, fragment_graph, replace_wire_cut_nodes


@batch_transform
def cut_circuit_mc(
    tape: QuantumTape,
    classical_processing_fn: Optional[callable] = None,
    auto_cutter: Union[bool, Callable] = False,
    max_depth: int = 1,
    shots: Optional[int] = None,
    device_wires: Optional[Wires] = None,
    **kwargs,
) -> Tuple[Tuple[QuantumTape], Callable]:
    """
    Cut up a circuit containing sample measurements into smaller fragments using a
    Monte Carlo method.

    Following the approach of `Peng et al. <https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.125.150504>`__,
    strategic placement of :class:`~.WireCut` operations can allow a quantum circuit to be split
    into disconnected circuit fragments. A circuit containing sample measurements can be cut and
    processed using Monte Carlo (MC) methods. This transform employs MC methods to allow for sampled measurement
    outcomes to be recombined to full bitstrings and, if a classical processing function is supplied,
    an expectation value will be evaluated.

    Args:
        tape (QuantumTape): the tape of the full circuit to be cut
        classical_processing_fn (callable): A classical postprocessing function to be applied to
            the reconstructed bitstrings. The expected input is a bitstring; a flat array of length ``wires``,
            and the output should be a single number within the interval :math:`[-1, 1]`.
            If not supplied, the transform will output samples.
        auto_cutter (Union[bool, Callable]): Toggle for enabling automatic cutting with the default
            :func:`~.kahypar_cut` partition method. Can also pass a graph partitioning function that
            takes an input graph and returns a list of edges to be cut based on a given set of
            constraints and objective. The default :func:`~.kahypar_cut` function requires KaHyPar to
            be installed using ``pip install kahypar`` for Linux and Mac users or visiting the
            instructions `here <https://kahypar.org>`__ to compile from source for Windows users.
        max_depth (int): The maximum depth used to expand the circuit while searching for wire cuts.
            Only applicable when transforming a QNode.
        shots (int): Number of shots. When transforming a QNode, this argument is
            set by the device's ``shots`` value or at QNode call time (if provided).
            Required when transforming a tape.
        device_wires (Wires): Wires of the device that the cut circuits are to be run on.
            When transforming a QNode, this argument is optional and will be set to the
            QNode's device wires. Required when transforming a tape.
        kwargs: Additional keyword arguments to be passed to a callable ``auto_cutter`` argument.
            For the default KaHyPar cutter, please refer to the docstring of functions
            :func:`~.find_and_place_cuts` and :func:`~.kahypar_cut` for the available arguments.

    Returns:
        Callable: Function which accepts the same arguments as the QNode.
        When called, this function will sample from the partitioned circuit fragments
        and combine the results using a Monte Carlo method.

    **Example**

    The following :math:`3`-qubit circuit contains a :class:`~.WireCut` operation and a :func:`~.sample`
    measurement. When decorated with ``@qml.cut_circuit_mc``, we can cut the circuit into two
    :math:`2`-qubit fragments:

    .. code-block:: python

        dev = qml.device("default.qubit", wires=2, shots=1000)

        @qml.cut_circuit_mc
        @qml.qnode(dev)
        def circuit(x):
            qml.RX(0.89, wires=0)
            qml.RY(0.5, wires=1)
            qml.RX(1.3, wires=2)

            qml.CNOT(wires=[0, 1])
            qml.WireCut(wires=1)
            qml.CNOT(wires=[1, 2])

            qml.RX(x, wires=0)
            qml.RY(0.7, wires=1)
            qml.RX(2.3, wires=2)
            return qml.sample(wires=[0, 2])

    we can then execute the circuit as usual by calling the QNode:

    >>> x = 0.3
    >>> circuit(x)
    tensor([[1, 1],
            [0, 1],
            [0, 1],
            ...,
            [0, 1],
            [0, 1],
            [0, 1]], requires_grad=True)

    Furthermore, the number of shots can be temporarily altered when calling
    the qnode:

    >>> results = circuit(x, shots=123)
    >>> results.shape
    (123, 2)

    Alternatively, if the optimal wire-cut placement is unknown for an arbitrary circuit, the
    ``auto_cutter`` option can be enabled to make attempts in finding such a optimal cut. The
    following examples shows this capability on the same circuit as above but with the
    :class:`~.WireCut` removed:

    .. code-block:: python

        @qml.cut_circuit_mc(auto_cutter=True)
        @qml.qnode(dev)
        def circuit(x):
            qml.RX(0.89, wires=0)
            qml.RY(0.5, wires=1)
            qml.RX(1.3, wires=2)

            qml.CNOT(wires=[0, 1])
            qml.CNOT(wires=[1, 2])

            qml.RX(x, wires=0)
            qml.RY(0.7, wires=1)
            qml.RX(2.3, wires=2)
            return qml.sample(wires=[0, 2])

    >>> results = circuit(x, shots=123)
    >>> results.shape
    (123, 2)


    .. details::
        :title: Usage Details

        Manually placing :class:`~.WireCut` operations and decorating the QNode with the
        ``cut_circuit_mc()`` batch transform is the suggested entrypoint into sampling-based
        circuit cutting using the Monte Carlo method. However,
        advanced users also have the option to work directly with a :class:`~.QuantumTape` and
        manipulate the tape to perform circuit cutting using the below functionality:

        .. autosummary::
            :toctree:

            ~transforms.qcut.tape_to_graph
            ~transforms.qcut.find_and_place_cuts
            ~transforms.qcut.replace_wire_cut_nodes
            ~transforms.qcut.fragment_graph
            ~transforms.qcut.graph_to_tape
            ~transforms.qcut.expand_fragment_tapes_mc
            ~transforms.qcut.qcut_processing_fn_sample
            ~transforms.qcut.qcut_processing_fn_mc

        The following shows how these elementary steps are combined as part of the
        ``cut_circuit_mc()`` transform.

        Consider the circuit below:

        .. code-block:: python

            np.random.seed(42)

            with qml.tape.QuantumTape() as tape:
                qml.Hadamard(wires=0)
                qml.CNOT(wires=[0, 1])
                qml.PauliX(wires=1)
                qml.WireCut(wires=1)
                qml.CNOT(wires=[1, 2])
                qml.sample(wires=[0, 1, 2])

        >>> print(tape.draw())
        0: ──H─╭●───────────┤ ╭Sample
        1: ────╰X──X──//─╭●─┤ ├Sample
        2: ──────────────╰X─┤ ╰Sample

        To cut the circuit, we first convert it to its graph representation:

        >>> graph = qml.transforms.qcut.tape_to_graph(tape)

        If, however, the optimal location of the :class:`~.WireCut` is unknown, we can use
        :func:`~.find_and_place_cuts` to make attempts in automatically finding such a cut
        given the device constraints. Using the same circuit as above but with the
        :class:`~.WireCut` removed, a slightly different cut with identical cost can be discovered
        and placed into the circuit with automatic cutting:

        .. code-block:: python

            with qml.tape.QuantumTape() as uncut_tape:
                qml.Hadamard(wires=0)
                qml.CNOT(wires=[0, 1])
                qml.PauliX(wires=1)
                qml.CNOT(wires=[1, 2])
                qml.sample(wires=[0, 1, 2])

        >>> cut_graph = qml.transforms.qcut.find_and_place_cuts(
        ...     graph=qml.transforms.qcut.tape_to_graph(uncut_tape),
        ...     cut_strategy=qml.transforms.qcut.CutStrategy(max_free_wires=2),
        ... )
        >>> print(qml.transforms.qcut.graph_to_tape(cut_graph).draw())
         0: ──H─╭●───────────┤  Sample[|1⟩⟨1|]
         1: ────╰X──//──X─╭●─┤  Sample[|1⟩⟨1|]
         2: ──────────────╰X─┤  Sample[|1⟩⟨1|]

        Our next step, using the original manual cut placement, is to remove the :class:`~.WireCut`
        nodes in the graph and replace with :class:`~.MeasureNode` and :class:`~.PrepareNode` pairs.

        >>> qml.transforms.qcut.replace_wire_cut_nodes(graph)

        The :class:`~.MeasureNode` and :class:`~.PrepareNode` pairs are placeholder operations that
        allow us to cut the circuit graph and then randomly select measurement and preparation
        configurations at cut locations. First, the :func:`~.fragment_graph` function pulls apart
        the graph into disconnected components as well as returning the
        `communication_graph <https://en.wikipedia.org/wiki/Quotient_graph>`__
        detailing the connectivity between the components.

        >>> fragments, communication_graph = qml.transforms.qcut.fragment_graph(graph)

        We now convert the ``fragments`` back to :class:`~.QuantumTape` objects

        >>> fragment_tapes = [qml.transforms.qcut.graph_to_tape(f) for f in fragments]

        The circuit fragments can now be visualized:

        >>> print(fragment_tapes[0].draw())
        0: ──H─╭●─────────────────┤  Sample[|1⟩⟨1|]
        1: ────╰X──X──MeasureNode─┤

        >>> print(fragment_tapes[1].draw())
        1: ──PrepareNode─╭●─┤  Sample[|1⟩⟨1|]
        2: ──────────────╰X─┤  Sample[|1⟩⟨1|]

        Additionally, we must remap the tape wires to match those available on our device.

        >>> dev = qml.device("default.qubit", wires=2, shots=1)
        >>> fragment_tapes = [qml.map_wires(t, dict(zip(t.wires, dev.wires))) for t in fragment_tapes]

        Note that the number of shots on the device is set to :math:`1` here since we
        will only require one execution per fragment configuration. In the
        following steps we introduce a shots value that will determine the number
        of fragment configurations. When using the ``cut_circuit_mc()`` decorator
        with a QNode, this shots value is automatically inferred from the provided
        device.

        Next, each circuit fragment is randomly expanded over :class:`~.MeasureNode` and
        :class:`~.PrepareNode` configurations. For each pair, a measurement is sampled from
        the Pauli basis and a state preparation is sampled from the corresponding pair of eigenstates.

        A settings array is also given which tracks the configuration pairs. Since each of the 4
        measurements has 2 possible eigenvectors, all configurations can be uniquely identified by
        8 values. The number of rows is determined by the number of cuts and the number of columns
        is determined by the number of shots.

        >>> shots = 3
        >>> configurations, settings = qml.transforms.qcut.expand_fragment_tapes_mc(
        ...     fragment_tapes, communication_graph, shots=shots
        ... )
        >>> tapes = tuple(tape for c in configurations for tape in c)
        >>> settings
        tensor([[6, 3, 4]], requires_grad=True)

        Each configuration is drawn below:

        >>> for t in tapes:
        ...     print(qml.drawer.tape_text(t))
        ...     print("")

        .. code-block::

            0: ──H─╭●────┤  Sample[|1⟩⟨1|]
            1: ────╰X──X─┤  Sample[Z]

            0: ──H─╭●────┤  Sample[|1⟩⟨1|]
            1: ────╰X──X─┤  Sample[X]

            0: ──H─╭●────┤  Sample[|1⟩⟨1|]
            1: ────╰X──X─┤  Sample[Y]

            0: ──I─╭●─┤  Sample[|1⟩⟨1|]
            1: ────╰X─┤  Sample[|1⟩⟨1|]

            0: ──X──S─╭●─┤  Sample[|1⟩⟨1|]
            1: ───────╰X─┤  Sample[|1⟩⟨1|]

            0: ──H─╭●─┤  Sample[|1⟩⟨1|]
            1: ────╰X─┤  Sample[|1⟩⟨1|]

        The last step is to execute the tapes and postprocess the results using
        :func:`~.qcut_processing_fn_sample`, which processes the results to approximate the original full circuit
        output bitstrings.

        >>> results = qml.execute(tapes, dev, gradient_fn=None)
        >>> qml.transforms.qcut.qcut_processing_fn_sample(
        ...     results,
        ...     communication_graph,
        ...     shots=shots,
        ... )
        [array([[0., 0., 0.],
                [1., 0., 0.],
                [1., 0., 0.]])]

        Alternatively, it is possible to calculate an expectation value if a classical
        processing function is provided that will accept the reconstructed circuit bitstrings
        and return a value in the interval :math:`[-1, 1]`:

        .. code-block::

            def fn(x):
                if x[0] == 0:
                    return 1
                if x[0] == 1:
                    return -1

        >>> qml.transforms.qcut.qcut_processing_fn_mc(
        ...     results,
        ...     communication_graph,
        ...     settings,
        ...     shots,
        ...     fn
        ... )
        array(-4.)

        Using the Monte Carlo approach of [Peng et. al](https://arxiv.org/abs/1904.00102), the
        `cut_circuit_mc` transform also supports returning sample-based expectation values of
        observables that are diagonal in the computational basis, as shown below for a `ZZ` measurement
        on wires `0` and `2`:

        .. code-block::

            dev = qml.device("default.qubit", wires=2, shots=10000)

            def observable(bitstring):
                return (-1) ** np.sum(bitstring)

            @qml.cut_circuit_mc(classical_processing_fn=observable)
            @qml.qnode(dev)
            def circuit(x):
                qml.RX(0.89, wires=0)
                qml.RY(0.5, wires=1)
                qml.RX(1.3, wires=2)

                qml.CNOT(wires=[0, 1])
                qml.WireCut(wires=1)
                qml.CNOT(wires=[1, 2])

                qml.RX(x, wires=0)
                qml.RY(0.7, wires=1)
                qml.RX(2.3, wires=2)
                return qml.sample(wires=[0, 2])

        We can now approximate the expectation value of the observable using

        >>> circuit(x)
        tensor(-0.776, requires_grad=True)
    """
    # pylint: disable=unused-argument, too-many-arguments

    if len(tape.measurements) != 1:
        raise ValueError(
            "The Monte Carlo circuit cutting workflow only supports circuits "
            "with a single output measurement"
        )

    if not all(isinstance(m, SampleMP) for m in tape.measurements):
        raise ValueError(
            "The Monte Carlo circuit cutting workflow only supports circuits "
            "with sampling-based measurements"
        )

    for meas in tape.measurements:
        if meas.obs is not None:
            raise ValueError(
                "The Monte Carlo circuit cutting workflow only "
                "supports measurements in the computational basis. Please only specify "
                "wires to be sampled within qml.sample(), do not pass observables."
            )

    g = tape_to_graph(tape)

    if auto_cutter is True or callable(auto_cutter):

        cut_strategy = kwargs.pop("cut_strategy", None) or CutStrategy(
            max_free_wires=len(device_wires)
        )

        g = find_and_place_cuts(
            graph=g,
            cut_method=auto_cutter if callable(auto_cutter) else kahypar_cut,
            cut_strategy=cut_strategy,
            **kwargs,
        )

    replace_wire_cut_nodes(g)
    fragments, communication_graph = fragment_graph(g)
    fragment_tapes = [graph_to_tape(f) for f in fragments]
    fragment_tapes = [qml.map_wires(t, dict(zip(t.wires, device_wires))) for t in fragment_tapes]

    configurations, settings = expand_fragment_tapes_mc(
        fragment_tapes, communication_graph, shots=shots
    )

    tapes = tuple(tape for c in configurations for tape in c)

    if classical_processing_fn:
        return tapes, partial(
            qcut_processing_fn_mc,
            communication_graph=communication_graph,
            settings=settings,
            shots=shots,
            classical_processing_fn=classical_processing_fn,
        )

    return tapes, partial(
        qcut_processing_fn_sample, communication_graph=communication_graph, shots=shots
    )


def _cut_circuit_mc_expand(
    tape: QuantumTape,
    classical_processing_fn: Optional[callable] = None,
    max_depth: int = 1,
    shots: Optional[int] = None,
    device_wires: Optional[Wires] = None,
    auto_cutter: Union[bool, Callable] = False,
    **kwargs,
):
    """Main entry point for expanding operations in sample-based tapes until
    reaching a depth that includes :class:`~.WireCut` operations."""
    # pylint: disable=unused-argument, too-many-arguments
    return _qcut_expand_fn(tape, max_depth, auto_cutter)


cut_circuit_mc.expand_fn = _cut_circuit_mc_expand


@cut_circuit_mc.custom_qnode_wrapper
def qnode_execution_wrapper_mc(self, qnode, targs, tkwargs):
    """Here, we overwrite the QNode execution wrapper in order
    to replace execution variables"""

    transform_max_diff = tkwargs.pop("max_diff", None)
    tkwargs.setdefault("device_wires", qnode.device.wires)

    if "shots" in inspect.signature(qnode.func).parameters:
        raise ValueError(
            "Detected 'shots' as an argument of the quantum function to transform. "
            "The 'shots' argument name is reserved for overriding the number of shots "
            "taken by the device."
        )

    def _wrapper(*args, **kwargs):
        if tkwargs.get("shots", False):
            raise ValueError(
                "Cannot provide a 'shots' value directly to the cut_circuit_mc "
                "decorator when transforming a QNode. Please provide the number of shots in "
                "the device or when calling the QNode."
            )

        shots = kwargs.pop("shots", False)
        shots = shots or qnode.device.shots

        if shots is None:
            raise ValueError(
                "A shots value must be provided in the device "
                "or when calling the QNode to be cut"
            )

        qnode.construct(args, kwargs)
        tapes, processing_fn = self.construct(qnode.qtape, *targs, **tkwargs, shots=shots)

        interface = qnode.interface
        execute_kwargs = getattr(qnode, "execute_kwargs", {}).copy()
        max_diff = execute_kwargs.pop("max_diff", 2)
        max_diff = transform_max_diff or max_diff

        gradient_fn = getattr(qnode, "gradient_fn", qnode.diff_method)
        gradient_kwargs = getattr(qnode, "gradient_kwargs", {})

        if interface is None or not self.differentiable:
            gradient_fn = None

        execute_kwargs["cache"] = False

        res = qml.execute(
            tapes,
            device=qnode.device,
            gradient_fn=gradient_fn,
            interface=interface,
            max_diff=max_diff,
            override_shots=1,
            gradient_kwargs=gradient_kwargs,
            **execute_kwargs,
        )

        out = processing_fn(res)
        if isinstance(out, list) and len(out) == 1:
            return out[0]
        return out

    return _wrapper


MC_STATES = [
    _prep_zero_state,
    _prep_one_state,
    _prep_plus_state,
    _prep_minus_state,
    _prep_iplus_state,
    _prep_iminus_state,
    _prep_zero_state,
    _prep_one_state,
]


def _identity(wire):
    qml.sample(qml.Identity(wires=wire))


def _pauliX(wire):
    qml.sample(qml.PauliX(wires=wire))


def _pauliY(wire):
    qml.sample(qml.PauliY(wires=wire))


def _pauliZ(wire):
    qml.sample(qml.PauliZ(wires=wire))


MC_MEASUREMENTS = [
    _identity,
    _identity,
    _pauliX,
    _pauliX,
    _pauliY,
    _pauliY,
    _pauliZ,
    _pauliZ,
]


def expand_fragment_tapes_mc(
    tapes: Sequence[QuantumTape], communication_graph: MultiDiGraph, shots: int
) -> Tuple[List[QuantumTape], np.ndarray]:
    """
    Expands fragment tapes into a sequence of random configurations of the contained pairs of
    :class:`MeasureNode` and :class:`PrepareNode` operations.

    For each pair, a measurement is sampled from
    the Pauli basis and a state preparation is sampled from the corresponding pair of eigenstates.
    A settings array is also given which tracks the configuration pairs. Since each of the 4
    measurements has 2 possible eigenvectors, all configurations can be uniquely identified by
    8 values. The number of rows is determined by the number of cuts and the number of columns
    is determined by the number of shots.

    .. note::

        This function is designed for use as part of the sampling-based circuit cutting workflow.
        Check out the :func:`~.cut_circuit_mc` transform for more details.

    Args:
        tapes (Sequence[QuantumTape]): the fragment tapes containing :class:`MeasureNode` and
            :class:`PrepareNode` operations to be expanded
        communication_graph (nx.MultiDiGraph): the communication (quotient) graph of the fragmented
            full graph
        shots (int): number of shots

    Returns:
        Tuple[List[QuantumTape], np.ndarray]: the tapes corresponding to each configuration and the
        settings that track each configuration pair

    **Example**

    Consider the following circuit that contains a sample measurement:

    .. code-block:: python

        with qml.tape.QuantumTape() as tape:
            qml.Hadamard(wires=0)
            qml.CNOT(wires=[0, 1])
            qml.WireCut(wires=1)
            qml.CNOT(wires=[1, 2])
            qml.sample(wires=[0, 1, 2])

    We can generate the fragment tapes using the following workflow:

    >>> g = qml.transforms.qcut.tape_to_graph(tape)
    >>> qml.transforms.qcut.replace_wire_cut_nodes(g)
    >>> subgraphs, communication_graph = qml.transforms.qcut.fragment_graph(g)
    >>> tapes = [qml.transforms.qcut.graph_to_tape(sg) for sg in subgraphs]

    We can then expand over the measurement and preparation nodes to generate random
    configurations using:

    .. code-block:: python

        >>> configs, settings = qml.transforms.qcut.expand_fragment_tapes_mc(tapes, communication_graph, 3)
        >>> print(settings)
        [[1 6 2]]
        >>> for i, (c1, c2) in enumerate(zip(configs[0], configs[1])):
        ...     print(f"config {i}:")
        ...     print(c1.draw())
        ...     print("")
        ...     print(c2.draw())
        ...     print("")
        ...

        config 0:
        0: ──H─╭●─┤  Sample[|1⟩⟨1|]
        1: ────╰X─┤  Sample[Z]

        1: ──I─╭●─┤  Sample[|1⟩⟨1|]
        2: ────╰X─┤  Sample[|1⟩⟨1|]

        config 1:
        0: ──H─╭●─┤  Sample[|1⟩⟨1|]
        1: ────╰X─┤  Sample[Y]

        1: ──H──S─╭●─┤  Sample[|1⟩⟨1|]
        2: ───────╰X─┤  Sample[|1⟩⟨1|]

        config 2:
        0: ──H─╭●─┤  Sample[|1⟩⟨1|]
        1: ────╰X─┤  Sample[Y]

        1: ──X──H──S─╭●─┤  Sample[|1⟩⟨1|]
        2: ──────────╰X─┤  Sample[|1⟩⟨1|]

    """
    pairs = [e[-1] for e in communication_graph.edges.data("pair")]
    settings = np.random.choice(range(8), size=(len(pairs), shots), replace=True)

    meas_settings = {pair[0].id: setting for pair, setting in zip(pairs, settings)}
    prep_settings = {pair[1].id: setting for pair, setting in zip(pairs, settings)}

    all_configs = []
    for tape in tapes:
        frag_config = []
        for shot in range(shots):
            with AnnotatedQueue() as q:
                for op in tape.operations:
                    w = op.wires[0]
                    if isinstance(op, PrepareNode):
                        MC_STATES[prep_settings[op.id][shot]](w)
                    elif not isinstance(op, MeasureNode):
                        qml.apply(op)

                for meas in tape.measurements:
                    qml.apply(meas)
                for op in tape.operations:
                    if isinstance(op, MeasureNode):
                        meas_w = op.wires[0]
                        MC_MEASUREMENTS[meas_settings[op.id][shot]](meas_w)

            frag_config.append(QuantumScript.from_queue(q))

        all_configs.append(frag_config)

    return all_configs, settings
